/*************************   TIFFLOAD.C   ***************************
**  Purpose:      Load a TIFF file into a TIFFIMAGE data structure.
**  Author:       James D. Murray, Anaheim, CA  USA
**  Date:         September 1990
**  C Compilers:  ANSI
**  Validated on: MSC 6.0, Turbo C++ 1.0
**  Memory Model: Any
**
**  Read a TIFF file and load the information into a TIFFIMAGE
**  structure.  All information within a TIFF file is written to
**  the TIFFIMAGE structure with the exception of the image data
**  itself.
**
**  The main() function in this file also calls the function
**  TiffLook() to display the information found within the
**  TIFFIMAGE structure.
**
**  If TiffLoad() will be reading very large TIFF files, or TIFF
**  files with many images, then compile using the compact or
**  large memory models to help avoid failed calls to malloc().
**
**  Copyright (C) 1990 by The C Gazette.  All rights reserved.
**  May be used freely if author and publication are acknowledged.
*******************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tiff.h"

short (*fgetshort)(FILE *);  /* Pointers to endian read functions */
long  (*fgetlong)(FILE *);   /* Declared extern in tifflook.c     */

/*
**  A working example of how to call TiffLoad() and TiffLook().
**
**  Parameters - The usual.
**
**  Returns: 0 on success and non-zero on failure.
*/
int main(int argc, char *argv[])
{
    TIFFIMAGE *tiffimage;

    if (argc < 2) {
        fputs("\nUsage: TIFFLOOK [ filename.TIF ]\n", stderr);
        return(1);
    }

    /* Read a TIFF file, load the image, and look at it */
    if ((tiffimage = TiffLoad(argv[1])) != (TIFFIMAGE *) NULL)
        TiffLook(tiffimage);
    else
        return(-1);

    return(0);
}

/*
**  Open, read, load, and close a TIFF file.
**
**  Parameters: tifffile - Pointer to the name of a TIFF file.
**
**  Returns: A pointer to a valid TIFFIMAGE structure on success,
**           otherwise NULL on failure.
*/
TIFFIMAGE *TiffLoad(char *tifffile)
{
    char       tiffname[80];    /* Name of the TIFF file */
    FILE      *tifffp;          /* Pointer to the TIFF file */
    TIFFIMAGE *tiffimage;       /* The TIFF file structure */

    /* If no file extension, assume .TIF */
    strcpy(tiffname, tifffile);
    if (!strchr(tiffname, '.'))
        strcat(tiffname, ".TIF");

    /* Open the TIFF file */
    if ((tifffp = fopen(tiffname, "rb")) == (FILE *) NULL) {
        fprintf(stderr, "Cannot open %s for reading.\n", tiffname);
        return((TIFFIMAGE *) NULL);
    }

    /* Read the TIFF file into the TIFF structure */
    if ((tiffimage = ReadTiffFile(tifffp, tiffname)) == (TIFFIMAGE *) NULL) 
    {
        fclose(tifffp);
        return((TIFFIMAGE *) NULL);
    }

    /* Close the TIFF image file */
    if (fclose(tifffp) == EOF) {
        fprintf(stderr, "tifflook: Error closing TIFF file %s\n",
          tiffname);
        return((TIFFIMAGE *) NULL);
    }
    return(tiffimage);
}

/*
**  Load a TIFF file into a TIFFIMAGE structure.
**
**  Parameters: TiffFp    - an open file pointer to a TIFF image
**              TiffImage - an unallocated TIFFIMAGE structure
**              TiffName  - the name of the TIFF file
**
**  Returns: A pointer to a valid TIFFIMAGE structure on success,
**           otherwise NULL on failure.
*/
TIFFIMAGE *ReadTiffFile(FILE *TiffFp, char *TiffName)
{
    long offset;            /* Offset to next IFD */
    TIFFIMAGE *tiffimage;   /* TIFFIMAGE structure */
    TIFFIFD *ifd;           /* IFD structure */
    TIFFIFD *current_ifd;   /* Current node of linked list */
    TIFFIFD *head;          /* Head of linked list */

    /* Allocate TIFF structure */
    if ((tiffimage =
        (TIFFIMAGE *) malloc(sizeof(TIFFIMAGE))) ==
        (TIFFIMAGE *) NULL) {
        fputs("tifflook: Cannot allocate memory for TIFF structure.\n",
          stderr);
        return((TIFFIMAGE *) NULL);
    }

    /* Initialize TIFF structure */
    strncpy(tiffimage->TIFF_FileName, TiffName, 80);

    if ((tiffimage->Header = ReadTiffIFH(TiffFp)) == (TIFFIFH *) NULL) 
    {
        fprintf(stderr, "tifflook: File %s is not a valid TIFF file.\n",
          TiffName);
        return((TIFFIMAGE *) NULL);
    }

    /*
    ** Read and store the IFD information of a TIFF file.  If there
    ** are multiple IFD's (images) then they are linked together.
    */
    offset = tiffimage->Header->IFDOffset;  /* Offset of 1st IFD */
    tiffimage->TIFF_NumOfImages = 0;        /* Clear image counter */
    head = (TIFFIFD *) NULL;                /* Clear the head pointer */
    do {
        if ((ifd = ReadTiffIFD(TiffFp, offset)) == (TIFFIFD *) NULL) 
        {
            fprintf
            ( stderr, "tifflook: Failed to load file %s at image %d.\n",
              TiffName, tiffimage->TIFF_NumOfImages );
            return((TIFFIMAGE *) NULL);
        }
        if (head == (TIFFIFD *) NULL) {     /* Singly-linked list */
            head = ifd;
            current_ifd = head;
        } else {
            current_ifd->Next = ifd;
            ifd->Next = (TIFFIFD *) NULL;
            current_ifd = ifd;
        }
        offset = ifd->NextIFDOffset;      /* Get offset to next IFD */
        tiffimage->TIFF_NumOfImages++;    /* Increment image counter */
    } while (offset);

    tiffimage->IfdList = head;            /* Attach IFD list to structure */

    return(tiffimage);                    /* Return completed structure */
}

/*
**  Read a TIFF Image File Header
**
**  Parameters: TiffFp - A pointer to the TIFF file.
**
**  Returns: A pointer to a valid IFH structure on success,
**           otherwise NULL on failure.
*/
TIFFIFH *ReadTiffIFH(FILE *TiffFp)
{
    TIFFIFH *ifh;       /* Image File Header structure */

    /* Allocate TIFF IFH structure */
    if ((ifh = (TIFFIFH *) malloc(sizeof(TIFFIFH))) == (TIFFIFH *) NULL) 
    {
        fputs("tifflook: Cannot allocate memory for TIFF header.\n", stderr);
        return((TIFFIFH *) NULL);
    }

    /* Read the byte order from the TIFF file */
    ifh->ByteOrder  = (SHORT) fgetc(TiffFp) & 0xff;
    ifh->ByteOrder += (SHORT) (fgetc(TiffFp) & 0xff) << 8;

    /*
    ** Decide which file I/O functions to use.
    */
    switch (ifh->ByteOrder) {
        case TIFF_MOTOROLA:             /* Big-endian */
            fgetshort = getbigshort;
            fgetlong  = getbiglong;
            break;
        case TIFF_INTEL:                /* Little-endian */
            fgetshort = getlittleshort;
            fgetlong  = getlittlelong;
            break;
        default:
            fprintf(stderr, "tifflook: unknown data format: %x\n",
              ifh->ByteOrder);
            return((TIFFIFH *) NULL);
    }

    /*
    ** Get the TIFF version number.
    */
    ifh->Version = fgetshort(TiffFp);
    if (ifh->Version != TIFF_VERSION) {
        fprintf(stderr, 
                "tifflook: unknown TIFF version: %d\n", ifh->Version);
        return((TIFFIFH *) NULL);
    }

    /*
    ** Get the offset to the first IFD (if any).
    */
    ifh->IFDOffset = fgetlong(TiffFp);
    if (ifh->IFDOffset == 0L) {
        fputs("tifflook: TIFF file has no image(s)!\n", stderr);
        return((TIFFIFH *) NULL);
    }
    return(ifh);    /* Return a valid IFH pointer */
}

/*
**  Read a TIFF Image File Directory.
**
**  An IFD is a singly-linked list of TIFF tags.
**
**  Parameters: TiffFp    - Pointer to TIFF file
**              IfdOffset - Offset of IFD within the TIFF image file
**
**  Returns: A pointer to a valid TIFFIFD structure on success,
**           otherwise NULL on failure.
*/
TIFFIFD *ReadTiffIFD(FILE *TiffFp, long IfdOffset)
{
    SHORT    tag_count;     /* Number of tags in an IFD */
    TIFFIFD *ifd;           /* Image File Directory structure */
    TIFFTAG *tag;           /* TIFF tag structure */
    TIFFTAG *head;          /* Head of linked list */
    TIFFTAG *current_tag;   /* Current node of linked list */

    /* Allocate TIFF IFD structure */
    if ((ifd = (TIFFIFD *) malloc(sizeof(TIFFIFD))) == (TIFFIFD *) NULL) 
    {
        fputs
        ("tifflook: Cannot allocate memory for TIFF Image File Directory.\n",
          stderr);
        return((TIFFIFD *) NULL);
    }

    /* Seek to IFD */
    fseek(TiffFp, IfdOffset, SEEK_SET);

    /* Get the number of tags in the IFD */
    tag_count = ifd->NumDirEntries = fgetshort(TiffFp);

    /* Read each tag in the IFD */
    for (head = (TIFFTAG *) NULL; tag_count; tag_count--) 
    {
        if ((tag = ReadTiffTag(TiffFp)) == (TIFFTAG *) NULL) 
        {
            fprintf(stderr, "tifflook: Error reading tag.\n");
            return((TIFFIFD *) NULL);
        }
        if (head == (TIFFTAG *) NULL)      /* Singly-linked list */
        {
            head = tag;
            current_tag = head;
        } 
        else 
        {
            current_tag->Next = tag;
            tag->Next = (TIFFTAG *) NULL;
            current_tag = tag;
        }
    }
    ifd->TagList = head;                    /* Attach tag list to IFD */
    ifd->NextIFDOffset = fgetlong(TiffFp);  /* Get offset to the next IFD */

    return(ifd);                            /* Return a valid IFD pointer */
}

/*
**  Read a tag from a TIFF file and return a tag structure.
**
**  Paramateres: TiffFp - A pointer to a TIFF file.
**
**  Returns: A valid tag structure on success, otherwise NULL.
*/
TIFFTAG *ReadTiffTag(FILE *TiffFp)
{
    LONG      i;         /* Counter used to step through tag->TagLength */
    long      savepos;   /* Save the current FILE offset value */
    BYTE     *t_byte;    /* Index variable for byte-oriented data */
    ASCII    *t_ascii;   /* Index variable for ascii-oriented data */
    SHORT    *t_short;   /* Index variable for short-oriented data */
    LONG     *t_long;    /* Index variable for long-oriented data */
    RATIONAL *num;       /* Index variable for rational-oriented data */
    RATIONAL *den;       /* Index variable for rational-oriented data */
    TIFFTAG  *tag;       /* The tag structure */

    /* Allocate TIFFTAG structure */
    if ((tag = (TIFFTAG *) malloc(sizeof(TIFFTAG))) == (TIFFTAG *) NULL) 
    {
        fprintf
        (stderr, "tifflook: Cannot allocate memory for tag structure.\n");
        return((TIFFTAG *) NULL);
    }

    /* Read the tag values from the TIFF file */
    tag->TagID       = fgetshort(TiffFp);
    tag->TagType     = fgetshort(TiffFp);
    tag->TagLength   = fgetlong(TiffFp);
    tag->ValueOffset = fgetlong(TiffFp);
    tag->Value       = (void *) NULL;

    /*
    ** Read the tag data value.
    **
    ** If the tag data is four bytes or less in length the
    ** tag->ValueOffset variable will contain the actual tag data.
    ** If the tag data is greater than four bytes in length, then
    ** tag->ValueOffset will contain a four-byte offset value to the
    ** location of the data within the TIFF file.  Note that data of
    ** type RATIONAL can never be found within a tag, as it takes
    ** 8 bytes to represent each RATIONAL value.
    */
    switch (tag->TagType) {
        case TYPE_BYTE:                 /* BYTE data */
            if (tag->TagLength < 5UL)
                return(tag);
            if (!(tag->Value =
                (BYTE *) calloc((size_t) tag->TagLength, sizeof(BYTE)))) 
            {
                fprintf
                (stderr, "tifflook: Cannot allocate memory for tag data\
                    (tagid: %d, size: %lu)\n",
                    tag->TagID, tag->TagLength * sizeof(BYTE));
                break;
            }
            savepos = ftell(TiffFp);
            fseek(TiffFp, tag->ValueOffset, SEEK_SET);
            t_byte = (BYTE *) tag->Value;
            for (i = 0; i < tag->TagLength; i++)
				*t_byte++ = (BYTE) fgetc(TiffFp);
            fseek(TiffFp, savepos, SEEK_SET);
            break;
        case TYPE_ASCII:                /* ASCII data */
            if (tag->TagLength < 5UL)
                return(tag);
            if (!(tag->Value =
                (ASCII *) calloc((size_t) tag->TagLength + 1, 
                                  sizeof(ASCII)))) 
            {
                fprintf
                (stderr, "tifflook: Cannot allocate memory for tag data\
                  (tagid: %d, size: %lu)\n",
                  tag->TagID, (tag->TagLength + 1) * sizeof(ASCII));
                break;
            }
            savepos = ftell(TiffFp);
            fseek(TiffFp, tag->ValueOffset, SEEK_SET);
            t_ascii = (ASCII *) tag->Value;
            for (i = 0; i <= tag->TagLength; i++)
				*t_ascii++ = (ASCII) fgetc(TiffFp);
            fseek(TiffFp, savepos, SEEK_SET);
            break;
        case TYPE_SHORT:                /* SHORT data */
            if (tag->TagLength < 3UL)
                return(tag);
            if (!(tag->Value =
                (SHORT *) calloc((size_t) tag->TagLength, sizeof(SHORT)))) 
            {
                fprintf
                (stderr, "tifflook: Cannot allocate memory for tag data\
                  (tagid: %d, size: %lu)\n",
                  tag->TagID, tag->TagLength * sizeof(SHORT));
                break;
            }
            savepos = ftell(TiffFp);
            fseek(TiffFp, tag->ValueOffset, SEEK_SET);
            t_short = tag->Value;
            for (i = 0; i < tag->TagLength; i++)
				*t_short++ = fgetshort(TiffFp);
            fseek(TiffFp, savepos, SEEK_SET);
            break;
        case TYPE_LONG:                 /* LONG data */
            if (tag->TagLength < 2UL)
                return(tag);
            if (!(tag->Value =
                (LONG *) calloc((size_t) tag->TagLength, sizeof(LONG)))) 
            {
                fprintf
                (stderr, "tifflook: Cannot allocate memory for tag data\
                  (tagid: %d, size: %lu)\n",
                  tag->TagID, tag->TagLength * sizeof(LONG));
                break;
            }
            savepos = ftell(TiffFp);
            fseek(TiffFp, tag->ValueOffset, SEEK_SET);
            t_long = tag->Value;
            for (i = 0; i < tag->TagLength; i++)
				*t_long++ = fgetlong(TiffFp);
            fseek(TiffFp, savepos, SEEK_SET);
            break;
        case TYPE_RATIONAL:             /* RATIONAL data */
            if (!(tag->Value =
                (RATIONAL *) calloc((size_t) tag->TagLength, 
                                    sizeof(RATIONAL) * 2))) 
            {
                fprintf
                (stderr, "tifflook: Cannot allocate memory for tag data\
                  (tagid: %d, size: %lu)\n",
                  tag->TagID, tag->TagLength * sizeof(RATIONAL));
                break;
            }
            savepos = ftell(TiffFp);
            fseek(TiffFp, tag->ValueOffset, SEEK_SET);
            num = (RATIONAL *) tag->Value;
            den = (RATIONAL *) tag->Value;
            for (i = 0, den++; i < tag->TagLength; i++, num++, den++) 
            {
				*num++ = fgetlong(TiffFp);
				*den++ = fgetlong(TiffFp);
            }
            fseek(TiffFp, savepos, SEEK_SET);
            break;
        default:
            fprintf(stderr, "tifflook: Unknown tag data type: %d\n",
              tag->TagType);
            return((TIFFTAG *) NULL);
    }
    return(tag);        /* Return a pointer to a valid tag structure */
}

/*
**  Read a short (16-bit) value in little-endian order.
**
**  Parameter: fp - A file stream to read data.
**
**  Returns: A short word value in little-endian order.
*/
short getlittleshort(FILE *fp)
{
    register short word;

    word  =  (fgetc(fp) & 0xff);
    word |= ((fgetc(fp) & 0xff) << 0x08);

    return(word);
}

/*
**  Read a short (16-bit) value in big-endian order.
**
**  Parameter: fp - A file stream to read data.
**
**  Returns: A short word value in big-endian order.
*/
short getbigshort(FILE *fp)
{
    register short word;

    word = (fgetc(fp) & 0xff);
    word = (fgetc(fp) & 0xff) | (word << 0x08);

    return(word);
}

/*
**  Read a long (32-bit) value in big-endian order.
**
**  Parameter: fp - A file stream to read data.
**
**  Returns: A long word value in big-endian order.
*/
long getbiglong(FILE *fp)
{
    register long word;

    word =  (long) (fgetc(fp) & 0xff);
    word = ((long) (fgetc(fp) & 0xff)) | (word << 0x08);
    word = ((long) (fgetc(fp) & 0xff)) | (word << 0x08);
    word = ((long) (fgetc(fp) & 0xff)) | (word << 0x08);

    return(word);
}

/*
**  Read a long (32-bit) value in little-endian order.
**
**  Parameter: fp - A file stream to read data.
**
**  Returns: A long word value in little-endian order.
*/
long getlittlelong(FILE *fp)
{
    register long word;

    word  =   (long) (fgetc(fp) & 0xff);
    word |= (((long) (fgetc(fp) & 0xff)) << 0x08);
    word |= (((long) (fgetc(fp) & 0xff)) << 0x10);
    word |= (((long) (fgetc(fp) & 0xff)) << 0x18);

    return(word);
}